import pytest
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate.branches import (
    Keyspace,
    calculate_random_seeds,
    expand_branch_templates,
)


def test_expand_branch_template():
    source = [
        {"a": {"b": [1, 2], "c": 3, "d": [4, 5, 6], "e": [True, False]}},
        {"a": {"b": 10, "c": 30, "d": 40, "e": True}},
    ]
    result = expand_branch_templates(source)

    result = [collapse_nested_dict(r) for r in result]

    expected = [
        collapse_nested_dict(r)
        for r in [
            {"a": {"b": 1, "c": 3, "d": 4, "e": True}},
            {"a": {"b": 2, "c": 3, "d": 5, "e": True}},
            {"a": {"b": 1, "c": 3, "d": 6, "e": True}},
            {"a": {"b": 2, "c": 3, "d": 4, "e": True}},
            {"a": {"b": 1, "c": 3, "d": 5, "e": True}},
            {"a": {"b": 2, "c": 3, "d": 6, "e": True}},
            {"a": {"b": 10, "c": 30, "d": 40, "e": True}},
            {"a": {"b": 1, "c": 3, "d": 4, "e": False}},
            {"a": {"b": 2, "c": 3, "d": 5, "e": False}},
            {"a": {"b": 1, "c": 3, "d": 6, "e": False}},
            {"a": {"b": 2, "c": 3, "d": 4, "e": False}},
            {"a": {"b": 1, "c": 3, "d": 5, "e": False}},
            {"a": {"b": 2, "c": 3, "d": 6, "e": False}},
        ]
    ]
    assert sorted(result) == sorted(expected)


@pytest.mark.parametrize("seed_count", [0, 10, 100, 1000])
def test_calculate_random_seeds(seed_count):
    seeds = calculate_random_seeds(seed_count)
    assert len(seeds) == len(set(seeds))
    assert len(seeds) == seed_count


@pytest.mark.parametrize("seed_count", [0, 10, 100, 1000])
def test_calculate_random_seeds_existing(seed_count):
    existing = list(range(20))

    seeds = calculate_random_seeds(seed_count, existing)
    assert len(seeds) == len(set(seeds))
    assert len(seeds + existing) == len(set(seeds + existing))
    assert len(seeds) == seed_count


def test_keyspace_order():
    ## divide an integer range of 15 into five bins
    input_draws, random_seeds, foos = [list(range(i, i + 3)) for i in range(0, 9, 3)]
    branches = [{"foo": foo} for foo in foos]
    # Define the correct ordering of the keyspace:
    ordered_keys = [
        (input_draw, random_seed, branch)
        # Iterate over branches, then seeds, then draws
        for input_draw in input_draws
        for random_seed in random_seeds
        for branch in branches
    ]

    full_keyspace = Keyspace(
        branches=branches, keyspace={"input_draw": input_draws, "random_seed": random_seeds}
    )

    for test_key, reference_key in zip(full_keyspace, ordered_keys):
        assert test_key == reference_key



"""
==============================
Model specification management
==============================

"""
import typing
from pathlib import Path
from typing import Optional

import yaml
from loguru import logger
from vivarium.config_tree import ConfigurationKeyError
from vivarium.framework.artifact import parse_artifact_path_config
from vivarium.framework.configuration import (
    ConfigTree,
    ConfigurationError,
    build_model_specification,
)

from vivarium_cluster_tools.psimulate import COMMANDS

if typing.TYPE_CHECKING:
    # Cyclic import.
    from vivarium_cluster_tools.psimulate.branches import Keyspace


# Configuration key constants
INPUT_DATA_KEY = "input_data"
ARTIFACT_PATH_KEY = "artifact_path"
FULL_ARTIFACT_PATH_KEY = f"{INPUT_DATA_KEY}.{ARTIFACT_PATH_KEY}"
OUTPUT_DATA_KEY = "output_data"
RESULTS_DIRECTORY_KEY = "results_directory"


def parse(
    command: str,
    input_model_specification_path: Optional[Path],
    artifact_path: Optional[Path],
    model_specification_path: Path,
    results_root: Path,
    keyspace: "Keyspace",
) -> ConfigTree:
    if command in [COMMANDS.restart, COMMANDS.expand]:
        return build_model_specification(model_specification_path)
    if command == COMMANDS.load_test:
        return build_model_specification()

    model_specification = build_model_specification(input_model_specification_path)
    model_specification.configuration.update(
        {OUTPUT_DATA_KEY: {RESULTS_DIRECTORY_KEY: str(results_root)}},
        source=__file__,
    )

    artifact_path_is_cli_arg = artifact_path is not None
    artifact_path_in_keyspace = FULL_ARTIFACT_PATH_KEY in keyspace

    if artifact_path_is_cli_arg and artifact_path_in_keyspace:
        raise ConfigurationError(
            "Artifact path cannot be specified both in the branch specification file"
            " and as a command line argument.",
            str(artifact_path),
        )
    elif artifact_path_is_cli_arg:
        model_specification.configuration[INPUT_DATA_KEY].update(
            {ARTIFACT_PATH_KEY: str(artifact_path)}, source=__file__
        )
    else:
        # Artifact path (if any) comes from the model spec.
        # The framework does not require an artifact, log a debug message in case this
        # is unintentional...
        try:
            parse_artifact_path_config(model_specification.configuration)
        except ConfigurationKeyError:
            logger.debug(
                "No artifact detected in arguments or configuration. This may"
                " be intentional. If not, supply one with the `-i` flag or in"
                " the configuration yaml."
            )

    return model_specification


def persist(
    model_specification: ConfigTree,
    model_specification_path: Path,
) -> None:
    model_specification_path.write_text(
        yaml.dump(model_specification.to_dict(), sort_keys=False)
    )



import pandas as pd
import pytest

from vivarium_cluster_tools.psimulate.results.processing import (
    _concat_preserve_types,
    _concat_results,
)


@pytest.mark.parametrize(
    "data_types",
    [
        [0, 1, 2],
        [0, 0.0, 1, 2, 3],
        [0.0, 1.0, 1.987987],
        [0, "a", 1, "b", 1.2],
        ["a", "b", "c", "d"],
    ],
    ids=[
        "just ints",
        "ints and floats",
        "just floats",
        "ints, floats, and strings",
        "just strings",
    ],
)
def test_concat_preserve_types(data_types):
    df = pd.DataFrame([data_types])
    df2 = pd.DataFrame([[d * 2 for d in data_types]])

    result = pd.DataFrame(_concat_preserve_types([df, df2]))

    expected_dtypes = df.dtypes

    expected_shape = (df.shape[0] + df2.shape[0], df.shape[1])

    for c in df:
        assert (result[c] == pd.concat([df[c], df2[c]]).reset_index(drop=True)).all()

    assert result.shape == expected_shape
    assert result.dtypes.sort_index().equals(expected_dtypes)


@pytest.mark.parametrize(
    "data_types",
    [
        [0, 1, 2],
        [0, 0.0, 1, 2, 3],
        [0.0, 1.0, 1.987987],
        [0, "a", 1, "b", 1.2],
        ["a", "b", "c", "d"],
    ],
    ids=[
        "just ints",
        "ints and floats",
        "just floats",
        "ints, floats, and strings",
        "just strings",
    ],
)
def test_concat_results(data_types):
    columns = [chr(i) for i in range(ord("a"), ord("a") + len(data_types))]
    old = pd.DataFrame([data_types], columns=columns)
    new = pd.DataFrame([[d * 2 for d in data_types]], columns=columns)

    old["input_draw"] = old["random_seed"] = 0.0
    new["input_draw"] = new["random_seed"] = 1.0

    combined = _concat_results(old, [new])

    expected_dtypes = old.dtypes

    expected_shape = (old.shape[0] + new.shape[0], old.shape[1])

    for c in old:
        assert (combined[c] == pd.concat([new[c], old[c]]).reset_index(drop=True)).all()

    assert combined.shape == expected_shape
    assert combined.dtypes.sort_index().equals(expected_dtypes.sort_index())

    # now no existing results
    no_old_combined = _concat_results(pd.DataFrame(), [new, old])

    assert no_old_combined.equals(combined)



"""
================
Load Test Worker
================

RQ worker executable for doing load testing.

"""
import time
from traceback import format_exc

import numpy as np
import pandas as pd
from loguru import logger
from rq import get_current_job
from vivarium.framework.randomness import get_hash

from vivarium_cluster_tools.psimulate.environment import ENV_VARIABLES
from vivarium_cluster_tools.psimulate.jobs import JobParameters

LOAD_TEST_WORK_HORSE_IMPORT_PATH = f"{__name__}.work_horse"


def get_psimulate_test_dict():
    return {"sleep": sleep_test, "large_results": large_results_test}


def sleep_test(job_parameters: JobParameters) -> pd.DataFrame:
    min_sleep_time = 5
    max_sleep_time = 60
    rs = np.random.RandomState(seed=get_hash(f"sleep_test_{job_parameters.random_seed}"))
    sleep_time = rs.randint(min_sleep_time, max_sleep_time)

    logger.info(f"Sleeping for {sleep_time}s.")
    time.sleep(sleep_time)
    logger.info(f"Sleep test successful.")
    return pd.DataFrame(
        {"sleep_time": sleep_time}, index=pd.Index([job_parameters.random_seed], name="seed")
    )


def large_results_test(job_parameters: JobParameters) -> pd.DataFrame:
    time.sleep(30)
    np.random.seed(seed=get_hash(f"large_results_test_{job_parameters.random_seed}"))
    return pd.DataFrame(np.random.random(10_000_000).reshape((1_000_000, 10)))


def work_horse(job_parameters: dict) -> pd.DataFrame:
    node = f"{ENV_VARIABLES.HOSTNAME.value}"
    job = f"{ENV_VARIABLES.JOB_ID.value}:{ENV_VARIABLES.TASK_ID.value}"

    job_parameters = JobParameters(**job_parameters)

    test_type = job_parameters.extras["test_type"]
    test_runner = get_psimulate_test_dict()[test_type]

    logger.info(f"Launching new job {job} on {node}")
    logger.info(f"Starting job: {job_parameters}")
    try:
        return test_runner(job_parameters)
    except Exception:
        logger.exception("Unhandled exception in worker")
        job = get_current_job()
        job.meta["root_exception"] = format_exc()
        job.save_meta()
        raise
    finally:
        logger.info(f"Exiting job: {job_parameters}")



import pytest

from vivarium_cluster_tools.psimulate.redis_dbs.registry import RegistryManager


@pytest.mark.parametrize("num_queues", [1, 7, 10, 19, 37])
@pytest.mark.parametrize("num_jobs", [1, 10, 42, 1337, 2023])
def test_allocate_jobs(mocker, num_queues, num_jobs):
    # mock the QueueManager class
    mocker.patch("vivarium_cluster_tools.psimulate.redis_dbs.registry.QueueManager")
    queues = [(f"queue_{i}", i) for i in range(num_queues)]
    manager = RegistryManager(queues, 0)
    # create a list of dummy jobs based on a list of integers
    jobs = [{"job": i} for i in range(num_jobs)]
    # make an array of all the jobs in every queue
    jobs_by_queue = list(manager.allocate_jobs(jobs))
    # Check that each "layer" of jobs has strictly smaller-priority jobs than the next
    largest_queue_size = max(len(queue) for queue in jobs_by_queue)
    for i in range(largest_queue_size - 1):
        this_layer = {queue[i]["job"] for queue in jobs_by_queue if len(queue) > i}
        next_layer = {queue[i + 1]["job"] for queue in jobs_by_queue if len(queue) > i + 1}
        assert max(this_layer) < min(next_layer)



"""
================
psimulate Runner
================

The main process loop for `psimulate` runs.

"""
import math
from collections import defaultdict
from pathlib import Path
from time import sleep, time

import pandas as pd
from loguru import logger

from vivarium_cluster_tools import logs
from vivarium_cluster_tools.psimulate import (
    COMMANDS,
    branches,
    cluster,
    jobs,
    model_specification,
    paths,
    pip_env,
    redis_dbs,
    results,
    worker,
)
from vivarium_cluster_tools.vipin.perf_report import report_performance


def process_job_results(
    registry_manager: redis_dbs.RegistryManager,
    existing_outputs: pd.DataFrame,
    output_directory: Path,
    no_batch: bool,
) -> defaultdict:
    written_results = existing_outputs
    unwritten_results = []
    batch_size = 0 if no_batch else 200
    status = defaultdict(int)

    logger.info("Entering main processing loop.")
    start_time = time()
    try:
        while registry_manager.jobs_to_finish:
            sleep(5)
            unwritten_results.extend(registry_manager.get_results())

            if len(unwritten_results) > batch_size:
                written_results, unwritten_results = results.write_results_batch(
                    output_directory,
                    written_results,
                    unwritten_results,
                    batch_size,
                )

            status = registry_manager.update_and_report()
            logger.info(f"Unwritten results: {len(unwritten_results)}")
            logger.info(f"Elapsed time: {(time() - start_time)/60:.1f} minutes.")
    finally:
        batch_size = 500
        while unwritten_results:
            written_results, unwritten_results = results.write_results_batch(
                output_directory,
                written_results,
                unwritten_results,
                batch_size=batch_size,
            )
            logger.info(f"Unwritten results: {len(unwritten_results)}")
            logger.info(f"Elapsed time: {(time() - start_time) / 60:.1f} minutes.")
        return status


def load_existing_outputs(result_path: Path, restart: bool) -> pd.DataFrame:
    try:
        existing_outputs = pd.read_hdf(result_path)
    except FileNotFoundError:
        existing_outputs = pd.DataFrame()
    assert (
        existing_outputs.empty or restart
    ), "How do you have existing outputs on an initial run?"
    return existing_outputs


def report_initial_status(
    num_jobs_completed: int, existing_outputs: pd.DataFrame, keyspace: branches.Keyspace
) -> None:
    if num_jobs_completed:
        logger.info(
            f"{num_jobs_completed} of {len(keyspace)} jobs completed in previous run."
        )
    if num_jobs_completed != len(existing_outputs):
        extra_jobs_completed = num_jobs_completed - len(existing_outputs)
        logger.warning(
            f"There are {extra_jobs_completed} jobs from the previous run which would not have been created "
            "with the configuration saved with the run. That either means that code "
            "has changed between then and now or that the outputs or configuration data "
            "have been modified. This may represent a serious error so give it some thought."
        )


def try_run_vipin(log_path: Path) -> None:
    try:
        report_performance(
            input_directory=log_path, output_directory=log_path, output_hdf=False, verbose=1
        )
    except Exception as e:
        logger.warning(f"Performance reporting failed with: {e}")


def main(
    command: str,
    input_paths: paths.InputPaths,
    native_specification: cluster.NativeSpecification,
    redis_processes: int,
    no_batch: bool,
    extra_args: dict,
) -> None:
    logger.info("Validating cluster environment.")
    cluster.validate_cluster_environment()

    # Generate programmatic representation of the output directory structure
    output_paths = paths.OutputPaths.from_entry_point_args(
        command=command,
        input_artifact_path=input_paths.artifact,
        result_directory=input_paths.result_directory,
        input_model_spec_path=input_paths.model_specification,
    )
    logger.info("Setting up output directory and all subdirectories.")
    output_paths.touch()

    logger.info("Setting up logging to files.")
    # Start sending logs to a file now that it exists.
    logs.configure_main_process_logging_to_file(output_paths.logging_root)
    logger.info("Validating programming environment.")
    # Either write a requirements.txt with the current environment
    # or verify the current environment matches the prior environment
    # used when doing a restart.
    pip_env.validate(output_paths.environment_file)

    logger.info(
        "Parsing input arguments into model specification and branches and writing to disk."
    )
    # Parse the branches configuration into a parameter space
    # and a flat representation of all parameters to be run.
    keyspace = branches.Keyspace.from_entry_point_args(
        input_branch_configuration_path=input_paths.branch_configuration,
        keyspace_path=output_paths.keyspace,
        branches_path=output_paths.branches,
        extras=extra_args,
    )
    # Throw that into our output directory. The keyspace output is
    # a cartesian product representation of the parameter space and
    # branches is a flat representation with the product expanded out.
    keyspace.persist(output_paths.keyspace, output_paths.branches)

    # Parse the model specification and resolve the artifact path
    # and then write to the output directory.
    model_spec = model_specification.parse(
        command=command,
        input_model_specification_path=input_paths.model_specification,
        artifact_path=input_paths.artifact,
        model_specification_path=output_paths.model_specification,
        results_root=output_paths.root,
        keyspace=keyspace,
    )
    model_specification.persist(model_spec, output_paths.model_specification)

    logger.info("Loading existing outputs if present.")
    # Load in any existing partial outputs if present.
    existing_outputs = load_existing_outputs(
        result_path=output_paths.results,
        restart=command in [COMMANDS.restart, COMMANDS.expand],
    )

    logger.info("Parsing arguments into worker job parameters.")
    # Translate the keyspace into the list of jobs to actually run
    # after accounting for any partially present results.
    job_parameters, num_jobs_completed = jobs.build_job_list(
        command=command,
        model_specification_path=output_paths.model_specification,
        output_root=output_paths.root,
        keyspace=keyspace,
        existing_outputs=existing_outputs,
        extras=extra_args,
    )
    # Let the user know if something is fishy at this point.
    report_initial_status(num_jobs_completed, existing_outputs, keyspace)
    if len(job_parameters) == 0:
        logger.info("No jobs to run, exiting.")
        return
    else:
        logger.info(f"Found {len(job_parameters)} jobs to run.")

    num_workers = len(job_parameters)
    if extra_args["max_workers"]:
        num_workers = min(extra_args["max_workers"], num_workers)

    logger.info("Spinning up Redis DBs and connecting to main process.")
    # Spin up the job & result dbs and get back (hostname, port) pairs for all the dbs.
    redis_ports = redis_dbs.launch(
        num_processes=redis_processes,
        num_workers=num_workers,
        redis_logging_root=output_paths.logging_root,
    )
    # Spin up a unified interface to all the redis databases
    registry_manager = redis_dbs.RegistryManager(redis_ports, num_jobs_completed)
    logger.info("Enqueuing jobs on Redis queues.")
    # Distribute all the remaining jobs across the job queues
    # in the redis databases.
    registry_manager.enqueue(
        jobs=job_parameters, workhorse_import_path=worker.WORK_HORSE_PATHS[command]
    )
    # Generate a worker template that chooses a redis DB at random to connect to.
    # This should (approximately) evenly distribute the workers over the work.
    redis_urls = [f"redis://{hostname}:{port}" for hostname, port in redis_ports]
    worker_template = (
        f"import random\nredis_urls = {redis_urls}\nREDIS_URL = random.choice(redis_urls)\n\n"
    )
    # Dump the worker config to a file we can pass to the workers on startup.
    output_paths.worker_settings.write_text(worker_template)
    worker_launch_script = worker.build_launch_script(
        worker_settings_file=output_paths.worker_settings,
        worker_log_directory=output_paths.worker_logging_root,
    )
    logger.info(f"Submitting redis workers to the cluster.")
    # Start an rq worker for every job using the cluster scheduler, up to an
    # optional user-specified limit. The workers will start as soon as they get
    # scheduled and start looking for work. They run in burst mode which means
    # they shut down if they can't find anything to do. This means it's
    # critical that we put the jobs on the queue before the workers land,
    # otherwise they'll just show up and shut down.

    cluster.submit_worker_jobs(
        num_workers=num_workers,
        worker_launch_script=worker_launch_script,
        cluster_logging_root=output_paths.cluster_logging_root,
        native_specification=native_specification,
    )

    logger.info(
        "Entering monitoring and results processing loop. "
        f"Results will be written to {str(output_paths.root)}"
    )
    # Enter the main monitoring and processing loop, which will check on
    # all the queues periodically, report status updates, and gather
    # and write results when they are available.
    status = process_job_results(
        registry_manager=registry_manager,
        existing_outputs=existing_outputs,
        output_directory=output_paths.root,
        no_batch=no_batch,
    )

    # Spit out a performance report for the workers.
    try_run_vipin(output_paths.worker_logging_root)

    # Emit warning if any jobs failed
    if status["failed"] > 0:
        logger.warning(
            f"*** NOTE: There {'was' if status['failed'] == 1 else 'were'} "
            f"{status['failed']} failed job{'' if status['failed'] == 1 else 's'}. ***"
        )

    logger.info(
        f"{status['finished']} of {status['total']} jobs completed successfully. "
        f"Results written to: {str(output_paths.root)}"
    )



"""
=========
vipin CLI
=========

Command line interface for `vipin`.

.. click:: vivarium_cluster_tools.vipin.cli:vipin
   :prog: vipin
   :show-nested:

"""
import click

from vivarium_cluster_tools import logs
from vivarium_cluster_tools.vipin import perf_report


@click.command()
@click.argument("logs-directory", type=click.Path(exists=True, file_okay=False))
@click.option(
    "--result-directory",
    "-o",
    type=click.Path(exists=True, file_okay=False),
    help="The directory into which to write the summary of the parsed logs. "
    "Defaults to given logs directory if not given.",
)
@click.option(
    "--hdf/--csv", default=False, help="Choose hdf or csv for output data. Defaults to csv."
)
@click.option("-v", "verbose", count=True, help="Configure logging verbosity.")
def vipin(logs_directory, result_directory, hdf, verbose):
    """Get performance information from worker_logs from a ``psimulate`` command.

    Given a worker logs directory from a previous run, a summary csv will be
    created in the ``result_directory`` (which defaults to the given logs
    directory unless otherwise specified).
    """
    logs.configure_main_process_logging_to_terminal(verbose, process_name="vipin")
    if not result_directory:
        result_directory = logs_directory

    perf_report.report_performance(logs_directory, result_directory, hdf, verbose)



"""
==============
psimulate Jobs
==============

"""
from pathlib import Path
from typing import List, NamedTuple, Tuple

import numpy as np
import pandas as pd
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate import COMMANDS, branches


class JobParameters(NamedTuple):
    """Parameters for a single distributed simulation job."""

    model_specification: str
    branch_configuration: dict
    input_draw: int
    random_seed: int
    results_path: str
    extras: dict

    @property
    def shared(self) -> dict:
        """Parameters shared by all jobs in a psimulate run."""
        return {
            "model_specification": self.model_specification,
            "results_path": self.results_path,
        }

    @property
    def job_specific(self) -> dict:
        """Parameters that vary by job in a psimulate run."""
        return {
            "input_draw": self.input_draw,
            "random_seed": self.random_seed,
            **self.branch_configuration,
        }

    def to_dict(self) -> dict:
        # I will never understand why this is a private
        # method of named tuples.
        return self._asdict()

    def __repr__(self) -> str:
        return (
            f"({self.input_draw}, {self.random_seed}, "
            f"{self.model_specification}, {self.branch_configuration})"
        )


def build_job_list(
    command: str,
    model_specification_path: Path,
    output_root: Path,
    keyspace: branches.Keyspace,
    existing_outputs: pd.DataFrame,
    extras: dict,
) -> Tuple[List[dict], int]:
    jobs = []
    number_already_completed = 0

    if command in [COMMANDS.run, COMMANDS.restart, COMMANDS.expand]:
        for (input_draw, random_seed, branch_config) in keyspace:
            parameters = JobParameters(
                model_specification=str(model_specification_path),
                branch_configuration=branch_config,
                input_draw=int(input_draw),
                random_seed=int(random_seed),
                results_path=str(output_root),
                extras={},
            )

            if already_complete(parameters, existing_outputs):
                number_already_completed += 1
            else:
                jobs.append(parameters.to_dict())
    else:
        assert command == COMMANDS.load_test
        for i in range(extras["num_workers"]):
            parameters = JobParameters(
                model_specification=str(model_specification_path),
                branch_configuration={},
                input_draw=0,
                random_seed=i,
                results_path=str(output_root),
                extras={"test_type": extras["test_type"]},
            )
            jobs.append(parameters.to_dict())

    np.random.shuffle(jobs)
    return jobs, number_already_completed


def already_complete(job_parameters: JobParameters, existing_outputs: pd.DataFrame) -> bool:
    if existing_outputs.empty:
        return False

    job_parameter_list = collapse_nested_dict(job_parameters.branch_configuration)
    job_parameter_list.extend(
        [
            ("run_configuration.run_key.input_draw", job_parameters.input_draw),
            ("run_configuration.run_key.random_seed", job_parameters.random_seed),
        ]
    )

    mask = pd.Series(True, index=existing_outputs.index)
    for k, v in job_parameter_list:
        if isinstance(v, float):
            mask &= np.isclose(existing_outputs[k], v)
        else:
            mask &= existing_outputs[k] == v
    return np.any(mask)



from time import time

import pytest
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate.jobs import JobParameters
from vivarium_cluster_tools.psimulate.worker.vivarium_work_horse import (
    parameter_update_format,
    setup_sim,
)


def test_setup_sim(mocker):
    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.get_current_job",
        return_value=mocker.Mock(id="test_string"),
    )
    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.time",
        return_value=1337,
    )
    job_parameters = JobParameters(
        model_specification=None,
        branch_configuration={
            "input_data": {"artifact_path": "~/vivarium.yaml"},
            "time": {"end": {"year": 2020}},
        },
        input_draw=1,
        random_seed=2,
        results_path="~/tmp",
        extras={},
    )

    initial_job_params = job_parameters.branch_configuration.copy()
    update_dict = parameter_update_format(job_parameters)
    sim_config = setup_sim(job_parameters).configuration.to_dict()
    job_config = job_parameters.branch_configuration

    ## Check that for all nested key, value pairs in ref_dict,
    ## the test dict has that key with the same value.
    def compare_dicts(test_dict, ref_dict):
        for k, v in ref_dict.items():
            if isinstance(v, dict):
                compare_dicts(test_dict[k], v)
            else:
                assert test_dict[k] == v

    for test_dict in [sim_config, job_config]:
        for ref_dict in [initial_job_params, update_dict]:
            compare_dicts(test_dict, ref_dict)



"""
==============================
Results Processing CLI Options
==============================

Command line options for configuring results handling in psimulate runs.

"""
import click

with_no_batch = click.option(
    "--no-batch", is_flag=True, help="Don't batch results, write them as they come in."
)
with_no_cleanup = click.option(
    "--no-cleanup",
    is_flag=True,
    hidden=True,
    help="Hidden developer option, if flagged, don't automatically "
    "cleanup results directory on failure.",
)



"""
=================
Cluster Interface
=================

"""

import atexit
import os
import shutil
from pathlib import Path
from typing import NamedTuple, TextIO

from vivarium_cluster_tools.psimulate.environment import ENV_VARIABLES
from vivarium_cluster_tools.utilities import get_drmaa


def validate_cluster_environment() -> None:
    if "slurm" not in ENV_VARIABLES.HOSTNAME.value:
        raise RuntimeError("This tool must be run from the IHME cluster.")

    submit_host_marker = "slogin"
    if submit_host_marker in ENV_VARIABLES.HOSTNAME.value:
        raise RuntimeError("This tool must not be run from a submit host.")


class NativeSpecification(NamedTuple):
    job_name: str
    project: str
    queue: str
    peak_memory: str
    max_runtime: str

    # Class constant
    NUM_THREADS: int = 1

    def to_cli_args(self):
        return (
            f"-J {self.job_name} "
            f"-A {self.project} "
            f"-p {self.queue} "
            f"--mem={self.peak_memory*1024} "
            f"-t {self.max_runtime} "
            f"-c {self.NUM_THREADS}"
        )


def submit_worker_jobs(
    num_workers: int,
    worker_launch_script: TextIO,
    cluster_logging_root: Path,
    native_specification: NativeSpecification,
) -> None:
    drmaa = get_drmaa()
    s = drmaa.Session()
    s.initialize()
    jt = s.createJobTemplate()
    jt.workingDirectory = os.getcwd()
    jt.remoteCommand = shutil.which("sh")
    jt.args = [worker_launch_script.name]
    jt.outputPath = f":{str(cluster_logging_root / '%A.%a.log')}"
    jt.errorPath = f":{str(cluster_logging_root / '%A.%a.log')}"
    jt.jobEnvironment = {
        "LC_ALL": "en_US.UTF-8",
        "LANG": "en_US.UTF-8",
    }
    jt.joinFiles = True
    jt.nativeSpecification = native_specification.to_cli_args()

    job_ids = s.runBulkJobs(jt, 1, num_workers, 1)
    array_job_id = job_ids[0].split(".")[0]

    def kill_jobs() -> None:
        try:
            s.control(array_job_id, drmaa.JobControlAction.TERMINATE)
        # FIXME: Hack around issue where drmaa.errors sometimes doesn't
        #        exist.
        except Exception as e:
            if "already completing" in str(e) or "Invalid job" in str(e):
                # This is the case where all our workers have already shut down
                # on their own, which isn't actually an error.
                pass
            else:
                raise

    atexit.register(kill_jobs)
